"""Tools for planning and scheduling analysis."""

from statistics import mean
from typing import Callable, Union

from ragraph.edge import Edge
from ragraph.graph import Graph
from ragraph.node import Node
from serde import asdict, from_dict

from raplan.classes import (
    Component,
    Horizon,
    Maintenance,
    Project,
    Schedule,
    ScheduleItem,
    System,
    Task,
)

NUMBERS = {int, float, Union[int, float]}
CLASSES = Union[
    Component,
    Horizon,
    Maintenance,
    Project,
    Schedule,
    ScheduleItem,
    System,
    Task,
]


class ClassToKind:
    """Class to Kind conversion."""

    COMPONENT = "component"
    HORIZON = "horizon"
    MAINTENANCE = "maintenance"
    PROJECT = "project"
    SCHEDULE = "schedule"
    SCHEDULEITEM = "schedule_item"
    SYSTEM = "system"
    TASK = "task"


class KindToClass:
    """Kind to Class conversion."""

    COMPONENT = Component
    HORIZON = Horizon
    MAINTENANCE = Maintenance
    PROJECT = Project
    SCHEDULE = Schedule
    SCHEDULE_ITEM = ScheduleItem
    SYSTEM = System
    TASK = Task


def _to_node(obj: CLASSES) -> Node:
    """Convert any of the classes to a Node."""
    # Use asdict to store serialized data to support later recreation.
    name = getattr(obj, "name", None)
    uuid = getattr(obj, "uuid", None)
    if name is None:
        node_name = uuid
    elif uuid is None:
        node_name = name
    else:
        node_name = f"{uuid}.{name}"
    node = Node(
        name=node_name,
        uuid=uuid,
        labels=[] if name is None else [name],
        kind=getattr(ClassToKind, obj.__class__.__name__.upper()),
        annotations=dict(data=asdict(obj)),
    )
    node._weights = dict()  # Hard set to empty.
    for f in obj.__dataclass_fields__.values():
        if f.type in NUMBERS:
            node.weights[f.name] = getattr(obj, f.name)
    return node


def _from_node(node: Node) -> CLASSES:
    """Convert node back into class."""
    cls = getattr(KindToClass, node.kind.upper())
    obj = from_dict(cls, node.annotations.data)
    for k, v in node.weights.items():
        if hasattr(obj, k):
            setattr(obj, k, v)
    return obj


def get_maintenance_graph(project: Project, threshold: Union[int, float]) -> Graph:
    """Get a graph of maintenance tasks. Edges are generated with an adjacency value of
    of the threshold minus the time difference between tasks.

    Arguments:
        project: Scheduling project to generate a graph for.
        threshold: Maximum time difference. Tasks that are within this threshold will be
            assigned an adjacency edge.

    Returns:
        Graph of maintenance tasks.
    """
    graph = Graph(
        labels=[project.name] if project.name else [],
        kind=ClassToKind.PROJECT,
        annotations=dict(data=asdict(project)),
    )

    maintenance_nodes: list[Node] = []
    for system in project.systems:
        if system.uuid in graph.node_uuid_dict:
            s = graph[system.uuid]
        else:
            s = _to_node(system)
            graph.add_node(s)

        for component in system.components:
            if component.uuid in graph.node_uuid_dict:
                c = graph[component.uuid]
            else:
                c = _to_node(component)
                graph.add_node(c)

            graph.add_edge(Edge(source=c, target=s, kind="belongs_to"))
            graph.add_edge(Edge(source=s, target=c, kind="consists_of"))

            for maintenance in component.maintenance:
                if maintenance.uuid in graph.node_uuid_dict:
                    m = graph[maintenance.uuid]
                else:
                    m = _to_node(maintenance)
                    graph.add_node(m)

                # Add edges to both component and system.
                graph.add_edge(Edge(source=m, target=c, kind="maintenance"))
                graph.add_edge(Edge(source=c, target=m, kind="maintenance"))
                graph.add_edge(Edge(source=m, target=s, kind="maintenance"))
                graph.add_edge(Edge(source=s, target=m, kind="maintenance"))

                for mx in maintenance_nodes:
                    delta_t = abs(m.weights["time"] - mx.weights["time"])
                    if not (delta_t < threshold):
                        continue
                    weight = 1.0 / (delta_t + 1.0)

                    graph.add_edge(
                        Edge(
                            source=m,
                            target=mx,
                            kind="adjacency",
                            weights=dict(adjacency=weight),
                        )
                    )
                    graph.add_edge(
                        Edge(
                            source=mx,
                            target=m,
                            kind="adjacency",
                            weights=dict(adjacency=weight),
                        )
                    )

                maintenance_nodes.append(m)

    return graph


def project_from_graph(graph: Graph) -> Project:
    """Recreate a Project from a Graph. Utilizes the 'data' annotation to build base
    objects, and overrides it with data found in the graph constructs.

    Arguments:
        graph: Graph data.
    """
    project = _from_node(graph)  # It works, even though naming is odd.
    assert isinstance(project, Project)

    systems = []
    for s in filter(lambda n: n.kind == "system" and n.is_leaf, graph.nodes):
        system = _from_node(s)
        assert isinstance(system, System)

        components: list[Component] = []
        for ec in graph.edges_from(s):
            if (
                ec.kind != "consists_of"
                or ec.target.kind != "component"
                or not ec.target.is_leaf
            ):
                continue
            c = ec.target

            component = _from_node(c)
            assert isinstance(component, Component)

            maintenance: list[Maintenance] = []
            for em in graph.edges_from(c):
                if (
                    em.kind != "maintenance"
                    or em.target.kind != "maintenance"
                    or not em.target.is_leaf
                ):
                    continue
                m = em.target

                maintenance_item = _from_node(m)
                assert isinstance(maintenance_item, Maintenance)
                maintenance.append(maintenance_item)

            component.maintenance = maintenance
            components.append(component)

        system.components = components
        systems.append(system)

    project.systems = systems
    return project


def _sync_start_earliest(nodes: list[Node]) -> list[Union[int, float]]:
    min_start = min(getattr(_from_node(n), "time", 0) for n in nodes)
    return len(nodes) * [min_start]


def _sync_start_latest(nodes: list[Node]) -> list[Union[int, float]]:
    max_start = max(getattr(_from_node(n), "time", 0) for n in nodes)
    return len(nodes) * [max_start]


def _sync_start_mean(nodes: list[Node]) -> list[Union[int, float]]:
    mean_start = mean(getattr(_from_node(n), "time", 0) for n in nodes)
    return len(nodes) * [mean_start]


def _get_end(n: Node) -> Union[int, float]:
    return getattr(_from_node(n), "end", 0)  # Maintenance


def _get_start_from_end(n: Node, end: Union[int, float]) -> Union[int, float]:
    m = _from_node(n)
    assert isinstance(m, Maintenance)
    return end - m.task.duration


def _sync_end_earliest(nodes: list[Node]) -> list[Union[int, float]]:
    min_end = min(_get_end(n) for n in nodes)
    return [_get_start_from_end(n, min_end) for n in nodes]


def _sync_end_latest(nodes: list[Node]) -> list[Union[int, float]]:
    max_end = max(_get_end(n) for n in nodes)
    return [_get_start_from_end(n, max_end) for n in nodes]


def _sync_end_mean(nodes: list[Node]) -> list[Union[int, float]]:
    mean_end = mean(_get_end(n) for n in nodes)
    return [_get_start_from_end(n, mean_end) for n in nodes]


SCHEDULE_FN = dict(
    SYNC_START_EARLIEST=_sync_start_earliest,
    SYNC_START_LATEST=_sync_start_latest,
    SYNC_START_MEAN=_sync_start_mean,
    SYNC_END_EARLIEST=_sync_end_earliest,
    SYNC_END_LATEST=_sync_end_latest,
    SYNC_END_MEAN=_sync_end_mean,
)
"""Maintenance task scheduling functions."""


def process_clustered_maintenance(
    graph: Graph,
    schedule_fn: Callable[[list[Node]], list[Union[int, float]]] = SCHEDULE_FN[
        "SYNC_START_EARLIEST"
    ],
):
    """Check for maintenance nodes with parents and synchronize their times.

    Arguments:
        graph: Maintenance graph.
        schedule_fn: A method that takes a list of (maintenance) nodes and returns a new
            list with modified starting times to incorporate. It should not modify the
            nodes in-place.
    """
    for n in graph.nodes:
        if n.kind != "maintenance" or n.is_leaf:
            continue
        leafs = n.children
        times = schedule_fn(leafs)
        for i, time in enumerate(times):
            leaf = leafs[i]
            leaf.weights["time"] = time
            leaf.annotations.data["time"] = time



"""RaPlan I/O with Excel. For this functionality to work, the 'excel' extra should be
installed.
"""

import json
import uuid
from pathlib import Path
from typing import Any, Callable, Optional, Union, get_args

import serde
import serde.json

from raplan import distributions
from raplan.classes import Component, Horizon, Maintenance, Project, System, Task

try:
    import openpyxl
    from openpyxl import Workbook
    from openpyxl.cell.cell import Cell
    from openpyxl.styles.protection import Protection
    from openpyxl.utils.cell import get_column_letter
    from openpyxl.worksheet.hyperlink import Hyperlink
    from openpyxl.worksheet.table import Table, TableStyleInfo
    from openpyxl.worksheet.worksheet import Worksheet
except ImportError:
    raise ImportError(
        "Package 'openpyxl' not found, please install it yourself "
        "or via the 'excel' package extra of RaPlan."
    )

_TYPES_TO_SHEETS = {
    Project: "Projects",
    System: "Systems",
    Component: "Components",
    Maintenance: "Maintenance",
    Task: "Tasks",
    Horizon: "Horizons",
}
_SHEETS_TO_TYPES = {v: k for k, v in _TYPES_TO_SHEETS.items()}
_FIELDS = dict()
for title, cls in _SHEETS_TO_TYPES.items():
    _FIELDS[title] = list(getattr(cls, "__dataclass_fields__", dict()).keys())
    _FIELDS[title].remove("uuid")  # Remove and add uuid as first column.
    _FIELDS[title].insert(0, "uuid")


def to_excel(
    projects: list[Project], path: Optional[Union[str, Path]]
) -> openpyxl.Workbook:
    """Convert a RaPlan project to an Excel workbook.

    Arguments:
        project: RaPlan project to export.
        path: Optional filename to save the workbook to.

    Returns:
        An openpyxl Workbook.
    """
    wb = Workbook()

    # Let schedule be the first sheet.
    schedule = wb.active
    schedule.title = "Schedule"
    schedule.append(
        [
            "project",
            "system",
            "component",
            "maintenance",
            "task",
            "maintenance_time",
            "task_rejuvenation",
            "task_duration",
            "task_cost",
        ]
    )
    _lock_worksheet(schedule)

    # Create sheets for instance storage.
    sheets: dict[str, Worksheet] = dict()
    indices: dict[str, dict[uuid.UUID, int]] = dict()
    objects: dict[uuid.UUID, Any] = dict()

    for cls, title in _TYPES_TO_SHEETS.items():  # Setup sheets, fields and headers.
        sheets[title] = wb.create_sheet(title)
        sheets[title].append(_FIELDS[title])
        indices[title] = dict()
        _lock_worksheet(sheets[title])

    for project in projects:  # Add data.
        horizon = project.horizon
        _add_obj_row(project, sheets, indices, objects, horizon)
        _add_obj_row(project.horizon, sheets, indices, objects, horizon)
        for system in project.systems:
            _add_obj_row(system, sheets, indices, objects, horizon)
            for component in system.components:
                _add_obj_row(component, sheets, indices, objects, horizon)
                for maintenance in component.maintenance:
                    _add_obj_row(maintenance, sheets, indices, objects, horizon)
                    _add_obj_row(maintenance.task, sheets, indices, objects, horizon)
                    _add_schedule_row(
                        schedule,
                        project,
                        system,
                        component,
                        maintenance,
                        wb,
                        indices,
                    )

    _set_defined_names(wb)  # Add local defined names for UUIDs.

    for ws in wb.worksheets:  # Set style and lock all worksheets.
        _set_table_style(ws)
        _set_auto_width(ws)

    if path:
        wb.save(str(Path(path)))

    return wb


def _add_obj_row(
    obj: Any,
    sheets: dict[str, Worksheet],
    indices: dict[str, dict[uuid.UUID, int]],
    objects: dict[uuid.UUID, Any],
    horizon: Horizon,
):
    key = _TYPES_TO_SHEETS[type(obj)]
    ws, fields = sheets[key], _FIELDS[key]
    if obj.uuid in objects:
        return
    row = [_get_cell_value(obj.uuid, sheets)]
    hyperlinks = []
    for field_index, field in enumerate(fields[1:]):
        value = getattr(obj, field, None)

        if "time" in field and value is not None and type(value) in {int, float}:
            value += horizon.start

        if isinstance(value, list):
            for iter_index, entry in enumerate(value):
                _ensure_header(ws, field, iter_index)
                row.append(_get_cell_value(entry, sheets))
                if type(entry) in _TYPES_TO_SHEETS:
                    # field_index + 2 because of for loop offset
                    hyperlinks.append(field_index + iter_index + 2)
            continue

        if type(value) in _TYPES_TO_SHEETS:
            hyperlinks.append(field_index + 2)

        row.append(_get_cell_value(value, sheets))

    ws.append(row)

    row_index = ws.max_row
    for col in hyperlinks:  # Make hyperlinks to defined names.
        _make_hyperlink(ws.cell(row_index, col))

    indices[key][obj.uuid] = row_index  # Add UUID to index list.
    objects[obj.uuid] = obj  # Add object to reference dict.


def _add_schedule_row(
    ws: Worksheet,
    project: Project,
    system: System,
    component: Component,
    maintenance: Maintenance,
    wb: Workbook,
    indices: dict[str, dict[uuid.UUID, int]],
):
    """
    project,
    system,
    component,
    maintenance,
    task,
    maintenance_time,
    task_rejuvenation,
    task_duration,
    task_cost,
    """
    ref_objs: list[Union[Project, System, Component, Maintenance, Task]] = [
        project,
        system,
        component,
        maintenance,
        maintenance.task,
    ]
    ref_uuids = [x.uuid for x in ref_objs]

    # Start with plain references
    row = [_encode_uuid(x) for x in ref_uuids]
    ws.append(row)
    for cell, obj in zip(ws[ws.max_row][: len(ref_objs)], ref_objs):
        cell.hyperlink = Hyperlink(
            ref="",
            location=cell.value,
        )
        if obj.name is not None:
            cell.value = obj.name

    objs: list[Union[Project, System, Component, Maintenance, Task]] = [
        maintenance,
        maintenance.task,
        maintenance.task,
        maintenance.task,
    ]
    props = [
        "time",
        "rejuvenation",
        "duration",
        "cost",
    ]
    src_row = ws.max_row
    src_col = len(row) + 1
    for obj, prop in zip(objs, props):
        _move_obj_value(obj, prop, ws.cell(src_row, src_col), wb, indices)
        src_col += 1


def _get_cell_value(value: Any, sheets: dict[str, Worksheet]) -> Any:
    if isinstance(value, uuid.UUID):
        return _encode_uuid(value)
    if isinstance(value, get_args(distributions.Distributions)):
        return serde.json.to_json({value.__class__.__name__: value})
    t = type(value)
    if t in _TYPES_TO_SHEETS:
        return _encode_uuid(value.uuid)
    return value


def _encode_uuid(value: uuid.UUID) -> str:
    return "_" + str(value).replace("-", "_")


def _ensure_header(ws: Worksheet, field: str, index: int):
    if index == 0:
        return
    columns = [c.value for c in ws[1]]
    name_index = f"{field}-{index}"
    if name_index in columns:
        return
    index_0 = columns.index(field) + 1
    ws.insert_cols(index_0 + 1)
    ws.cell(1, index_0 + 1, name_index)


def _move_obj_value(
    obj,
    field: str,
    target: Cell,
    wb: Workbook,
    indices: dict[str, dict[uuid.UUID, int]],
):
    key = _TYPES_TO_SHEETS[type(obj)]
    src_row = indices[key][obj.uuid]
    src_col = _FIELDS[key].index(field) + 1
    source = wb[key].cell(src_row, src_col)
    _move_value(wb, source, target, f"{key}_{field}{_encode_uuid(obj.uuid)}")


def _move_value(wb: Workbook, source: Cell, target: Cell, name: str):
    ref = f"{target.parent.title}!${target.column_letter}${target.row}"
    target.value = source.value
    try:
        _add_defined_name(wb, name, ref)
        _make_hyperlink(source, name)
        source.value = f"={ref}"
        _make_editable(target)
    except ValueError as e:
        # A task or maintenance object may have been re-used.
        if "already exists" in str(e):
            target.value = f"={ref}"
            _make_hyperlink(target, name)
        else:
            raise e


def _make_hyperlink(cell: Cell, name: Optional[str] = None):
    location = name or str(cell.value)
    cell.hyperlink = Hyperlink(ref="", location=location)


def _make_editable(cell: Cell):
    cell.protection = Protection(locked=False)


def _add_defined_name(wb: Workbook, name: str, ref: str):
    wb.defined_names.append(
        openpyxl.workbook.defined_name.DefinedName(
            name,
            attr_text=ref,
        )
    )


def _set_defined_names(wb: Workbook):
    for ws in wb.worksheets[1:]:
        for row in ws[f"A2:A{ws.max_row}"]:
            cell = row[0]
            column_letter = get_column_letter(ws.max_column)
            _add_defined_name(
                wb,
                cell.value,
                f"{ws.title}!$B${cell.row}:${column_letter}${cell.row}",
            )
        ws.column_dimensions["A"].hidden = True


def _set_table_style(ws: Worksheet):
    if ws.max_row < 2:  # openpyxl can't make a proper table without content.
        return
    ref = f"A1:{get_column_letter(ws.max_column)}{ws.max_row}"
    name = ws.title
    tab = Table(name=name, displayName=name, ref=ref)

    # Add a default style with striped rows and banded columns
    style = TableStyleInfo(
        name="TableStyleMedium9",
        showFirstColumn=False,
        showLastColumn=False,
        showRowStripes=True,
        showColumnStripes=False,
    )
    tab.tableStyleInfo = style
    ws.add_table(tab)


def _set_auto_width(ws: Worksheet):
    for i in range(1, ws.max_column + 1):
        letter = get_column_letter(i)
        ws.column_dimensions[letter].bestFit = True


def _lock_worksheet(ws: Worksheet):
    ws.protection.sheet = True
    ws.protection.password = "raplan"
    ws.protection.enable()


def from_excel(path: Union[str, Path]) -> list[Project]:
    """Convert an Excel file to a RaPlan project.

    Arguments:
        path: Path to an Excel workbook containing a RaPlan project.
    """
    wb = openpyxl.load_workbook(path)
    objects: dict[uuid.UUID, Any] = dict()
    for sheetname in [
        "Horizons",
        "Tasks",
        "Maintenance",
        "Components",
        "Systems",
    ]:
        ws = wb[sheetname]
        for i in range(2, ws.max_row + 1):
            _from_row(ws[i], sheetname, wb, objects)
    ws = wb["Projects"]
    projects = [
        _from_row(ws[i], "Projects", wb, objects) for i in range(2, ws.max_row + 1)
    ]
    return projects


def _decode_uuid(enc: str):
    return uuid.UUID(enc[1:].replace("_", "-"), version=4)


def _decode_distribution(enc: str):
    d = json.loads(enc)
    for k, v in d.items():
        return serde.from_dict(getattr(distributions, k), v)


_CELL_VALUE_CONVERSION: dict[Any, Callable[[Any], Any]] = {
    Union[int, float]: float,
    Optional[Union[int, float]]: float,
    Optional[str]: str,
    str: str,
    uuid.UUID: _decode_uuid,
    distributions.Distributions: _decode_distribution,
}


def _from_row(
    row: tuple[Cell, ...],
    sheetname: str,
    wb: Workbook,
    objects: dict[uuid.UUID, Any],
):
    cls = _SHEETS_TO_TYPES[sheetname]
    fields = _FIELDS[sheetname]
    kwargs = dict()
    for field in fields:
        value = _arg_from_row(row, field, sheetname, wb, objects)
        if value is None:
            continue
        kwargs[field] = value
    obj = cls(**kwargs)
    objects[obj.uuid] = obj
    return obj


def _arg_from_row(
    row: tuple[Cell, ...],
    field: str,
    sheetname: str,
    wb: Workbook,
    objects: dict[uuid.UUID, Any],
):
    cls = _SHEETS_TO_TYPES[sheetname]

    t = getattr(getattr(cls, "__dataclass_fields__", dict())[field], "type", None)
    index = _FIELDS[sheetname].index(field)
    value = _resolve_value(wb, row[index])
    if value == "" or value is None or t is None:
        return None

    if t in _CELL_VALUE_CONVERSION:
        return _CELL_VALUE_CONVERSION[t](value)

    if t in _TYPES_TO_SHEETS:
        return objects[_decode_uuid(value)]

    is_list = (
        getattr(
            getattr(cls, "__dataclass_fields__", dict())[field], "default_factory", None
        )
        == list
    )
    if is_list:
        headers = [c.value for c in wb[sheetname][1]]
        result = [objects[_decode_uuid(value)]]
        offset = 1
        while f"{field}-{offset}" in headers:
            enc_uuid = _resolve_value(wb, row[index + offset])
            if enc_uuid is None or enc_uuid == "":
                break
            result.append(objects[_decode_uuid(enc_uuid)])
            offset += 1
        return result

    return None


def _resolve_value(wb: Workbook, cell: Cell) -> str:
    if str(cell.value).startswith("="):
        parts = str(cell.value)[1:].split("!")
        if len(parts) < 2:
            return _resolve_value(wb, cell.parent[parts[0]])
        return _resolve_value(wb, wb[parts[0]][parts[1]])
    else:
        return cell.value



"""Dataclasses to use and configure planning and scheduling with."""

from dataclasses import dataclass
from typing import Optional, Union
from uuid import UUID, uuid4

from serde import InternalTagging, field, serde

from raplan import distributions

__all__ = [
    "Task",
    "Maintenance",
    "Component",
    "System",
    "Horizon",
    "Procedure",
    "Project",
    "ScheduleItem",
    "Schedule",
    "Procedure",
    "CyclicStrategy",
]


def _is_empty(lst: Union[list, dict, set]) -> bool:
    """Check whether a list is empty."""
    return len(lst) == 0


@serde(tagging=InternalTagging("type"))
@dataclass
class Task:
    """Maintenance task to apply to a component.

    Arguments:
        name: Name for this action.
        rejuvenation: Rejuvenation factor between [0.0-1.0]. Percentage of age that is
            regained. Therefore, 1.0 would mean a full replacement.
        duration: Duration of the maintenance. Usually in years.
        cost: Cost of the maintenance. Usually expressed in a currency or equivalent.
    """

    name: Optional[str] = field(default=None, skip_if_default=True)
    rejuvenation: Union[int, float] = 1.0
    duration: Union[int, float] = 1.0
    cost: Union[int, float] = 1.0
    uuid: UUID = field(default_factory=uuid4)


@serde(tagging=InternalTagging("type"))
@dataclass
class Maintenance:
    """Maintenance task scheduled at a point in time.

    Arguments:
        name: Name of this maintenance task.
        task: Task information.
        time: Time at which this maintenance is scheduled.
    """

    name: Optional[str] = field(default=None, skip_if_default=True)
    task: Task = field(default_factory=Task)
    time: Union[int, float] = 1.0
    uuid: UUID = field(default_factory=uuid4)

    @property
    def end(self) -> Union[int, float]:
        """End time of this maintenance."""
        return self.time + self.task.duration


@serde(tagging=InternalTagging("type"))
@dataclass
class Component:
    """Component with a failure distribution.

    Arguments:
        name: Name of this component.
        age: Starting age offset (usually in years).
        distribution: Failure distribution to use.
        maintenance: List of maintenance tasks that should be applied over this
            component's lifespan.
    """

    name: Optional[str] = field(default=None, skip_if_default=True)
    age: Union[int, float] = 0.0
    lifetime: Union[int, float] = 1.0
    distribution: distributions.Distributions = field(
        default_factory=distributions.Weibull
    )
    maintenance: list[Maintenance] = field(
        default_factory=list, skip_if=_is_empty, repr=False
    )
    uuid: UUID = field(default_factory=uuid4)

    def get_ordered_maintenance(self) -> list[Maintenance]:
        """Maintenance tasks sorted in time."""
        return sorted(self.maintenance, key=lambda m: m.time)

    def cfp(self, x: Union[int, float] = 1.0) -> float:
        """Cumulative failure probability density function incorporating maintenance."""
        return self.distribution.cdf(self.get_age_at(x))

    def get_age_at(self, x: Union[int, float] = 1.0) -> float:
        """Effective age at a point in time given the currently set schedule."""
        age = float(self.age)
        last_time = 0.0
        for m in self.get_ordered_maintenance():
            if m.time > x:
                # Maintenance is yet to happen.
                break
            # Apply rejuvenation with the then actual age.
            age = (age + m.time - last_time) * (1.0 - m.task.rejuvenation)
            last_time = m.time
        # Add remaining time since last maintenance.
        age += x - last_time
        return age

    def schedule_maintenance(self, maintenance: Maintenance):
        """Schedule maintenance for a single or all system's component or all
        components.

        Arguments:
            maintenance: Maintenance to schedule.
        """
        self.maintenance.append(maintenance)


@serde(tagging=InternalTagging("type"))
@dataclass
class System:
    """A system consisting of multiple components.

    Arguments:
        name: Name of this system.
        components: Components of this system.
    """

    name: Optional[str] = field(default=None, skip_if_default=True)
    components: list[Component] = field(
        default_factory=list, skip_if=_is_empty, repr=False
    )
    uuid: UUID = field(default_factory=uuid4)

    def cfp(self, x: Union[int, float] = 1.0) -> float:
        """Cumulative failure probability density function as the sum of its
        components' respective function incorporating maintenance.
        """
        if len(self.components):
            return distributions.compound_probability(c.cfp(x) for c in self.components)
        else:
            return 0.0

    def get_ordered_maintenance(self) -> list[Maintenance]:
        """Get all maintenance ordered in time."""
        return sorted(
            [m for c in self.components for m in c.maintenance], key=lambda m: m.time
        )

    def get_component(self, name: str) -> Component:
        """Get a component by name."""
        for c in self.components:
            if c.name == name:
                return c
        raise KeyError(f"Component with name '{name}' does not exist in this system.")

    def schedule_maintenance(
        self, maintenance: Maintenance, component: Optional[str] = None
    ):
        """Schedule maintenance for a single or all system's component or all
        components.

        Arguments:
            maintenance: Maintenance to schedule.
            component: Component name. If kept `None`, it will be applied to all.
        """
        if component is None:
            for c in self.components:
                c.schedule_maintenance(maintenance)
        else:
            self.get_component(component).schedule_maintenance(maintenance)


@serde(tagging=InternalTagging("type"))
@dataclass
class Horizon:
    """Planning and scheduling horizon.

    Arguments:
        start: Start of the planning horizon.
        end: End of the planning horizon. Optional, as it is otherwise derived from the
            final task in the schedule.
    """

    start: Union[int, float] = 0.0
    end: Optional[Union[int, float]] = field(default=None, skip_if_default=True)
    uuid: UUID = field(default_factory=uuid4)

    def get_range(self, steps: int, zero_based: bool = True) -> list[Union[int, float]]:
        """Range between start and end (inclusive) in the given number of steps."""
        if self.end is None:
            raise ValueError("Can't calculate a range with no horizon end value.")
        step_size = (self.end - self.start) / steps
        start = type(self.start)(0) if zero_based else self.start
        return [start + i * step_size for i in range(steps + 1)]


@serde(tagging=InternalTagging("type"))
@dataclass
class Project:
    """Planning and scheduling project."""

    name: Optional[str] = field(default=None, skip_if_default=True)
    horizon: Horizon = field(default_factory=Horizon)
    systems: list[System] = field(default_factory=list, skip_if=_is_empty, repr=False)
    uuid: UUID = field(default_factory=uuid4)

    def get_horizon_end(self) -> float:
        """Get the end of the planning horizon or last maintenance task."""
        if self.horizon.end is None:
            end = 0.0
            try:
                end = max(
                    m.time
                    for s in self.systems
                    for c in s.components
                    for m in c.maintenance
                )
            except ValueError:
                pass  # arg is an empty sequency: end = 0.0
            finally:
                return end
        return self.horizon.end

    def cfp(self, x: Union[int, float] = 1.0) -> float:
        """Cumulative failure probability density function as the sum of its
        systems' respective function incorporating maintenance.
        """
        if len(self.systems):
            return distributions.compound_probability(s.cfp(x) for s in self.systems)
        else:
            return 0.0

    def get_ordered_maintenance(self) -> list[Maintenance]:
        """Get all maintenance ordered in time."""
        return sorted(
            [m for s in self.systems for c in s.components for m in c.maintenance],
            key=lambda m: m.time,
        )

    def get_system(self, name: str) -> System:
        """Get a component by name."""
        for s in self.systems:
            if s.name == name:
                return s
        raise KeyError(f"System with name '{name}' does not exist in this project.")

    def schedule_maintenance(
        self,
        maintenance: Maintenance,
        system: Optional[str] = None,
        component: Optional[str] = None,
    ):
        """Schedule maintenance for a single or all system's component or all
        components.

        Arguments:
            maintenance: Maintenance to schedule.
            system: System name. If kept `None`, it will be applied to all.
            component: Component name. If kept `None`, it will be applied to all.
        """
        if system is None:
            for s in self.systems:
                s.schedule_maintenance(maintenance, component=component)
        else:
            self.get_system(system).schedule_maintenance(
                maintenance, component=component
            )

    def get_schedule(self) -> "Schedule":
        """Get a fully generated schedule."""
        return Schedule(
            items=[
                ScheduleItem(
                    name=m.task.name,
                    project=self.name,
                    system=s.name,
                    component=c.name,
                    maintenance=m.name,
                    task=m.task.name,
                    rejuvenation=m.task.rejuvenation,
                    duration=m.task.duration,
                    cost=m.task.cost,
                    time=m.time,
                )
                for s in self.systems
                for c in s.components
                for m in c.maintenance
            ]
        )


@serde(tagging=InternalTagging("type"))
@dataclass
class ScheduleItem:
    """A schedule item with full detail regarding its system, component and maintenance
    task info.

    Arguments:
        name: Name for this action. rejuvenation: Rejuvenation factor between [0.0-1.0].
        Percentage of age that is
            regained. Therefore, 1.0 would mean a full replacement.
        duration: Duration of the maintenance. Usually in years. cost: Cost of the
        maintenance. Usually expressed in a currency or equivalent. system: Name of the
        system to which this maintenance is applied. component: Name of the component to
        which this maintenance is applied. time: Time at which this maintenance is
        scheduled.
    """

    name: Optional[str] = field(default=None, skip_if_default=True)
    project: Optional[str] = field(default=None, skip_if_default=True)
    system: Optional[str] = field(default=None, skip_if_default=True)
    component: Optional[str] = field(default=None, skip_if_default=True)
    maintenance: Optional[str] = field(default=None, skip_if_default=True)
    task: Optional[str] = field(default=None, skip_if_default=True)
    rejuvenation: Union[int, float] = 1.0
    duration: Union[int, float] = 1.0
    cost: Union[int, float] = 1.0
    time: Union[int, float] = 1.0
    uuid: UUID = field(default_factory=uuid4)


@serde(tagging=InternalTagging("type"))
@dataclass
class Schedule:
    """A full maintenance schedule.

    Arguments:
        items: Scheduled tasks.
    """

    items: list[ScheduleItem] = field(default_factory=list, skip_if=_is_empty)

    def get_ordered_maintenance(self) -> list[ScheduleItem]:
        """Get all tasks ordered in time."""
        return sorted(self.items, key=lambda t: t.time)

    @classmethod
    def from_projects(cls, projects: list[Project]) -> "Schedule":
        """Create a schedule for multiple projects."""

        schedules = [p.get_schedule() for p in projects]
        return cls(items=[i for s in schedules for i in s.items])


@serde(tagging=InternalTagging("type"))
@dataclass
class Procedure:
    """A specific grouping of tasks to apply to a System.

    Note:
        Mainly used for plotting purposes.

    Arguments:
        name: Name for this procedure.
        system: System (name) to which the procedure should be applied.
        kind: Kind of procedure (category name).
        time: Time at which the procedure is scheduled.
        cost: Cost of the procedure.
        duration: Duration of the procedure.
    """

    name: Optional[str] = field(default=None, skip_if_default=True)
    system: str = "system"
    kind: str = "procedure"
    time: Union[int, float] = 1.0
    cost: Union[int, float] = 1.0
    duration: Union[int, float] = 1.0
    uuid: UUID = field(default_factory=uuid4)


@serde(tagging=InternalTagging("type"))
@dataclass
class CyclicStrategy:
    """Maintenance strategy to renovate or replace a component at certain percentages
    of a cycle.
    """

    tasks: list[Task] = field(default_factory=list, skip_if=_is_empty)
    percentages: list[float] = field(default_factory=list, skip_if=_is_empty)

    def apply_to_component(
        self,
        component: Component,
        cycle_length: Union[int, float],
        horizon: Horizon,
        repeat: bool = True,
        include_history: bool = True,
        integers: bool = False,
        overwrite: bool = True,
    ) -> None:
        """Apply this strategy to a component.

        Arguments:
            component: Component for which to schedule maintenance.
            cycle_length: Cycle length.
            horizon: Planning horizon to consider.
            repeat: Whether the cycle should be repeated until the end of the horizon.
            include_history: Whether to include historical maintenance entries for
                components that have a pre-defined age.
            overwrite: Whether to fully overwrite a component's maintenance planning
                with this new one or extend it.
            integers: Whether to force all times to be integers.
        """
        maintenance = self.get_maintenance(
            component.age,
            cycle_length,
            horizon,
            repeat=repeat,
            include_history=include_history,
            integers=integers,
        )
        if overwrite:
            component.maintenance = maintenance
        else:
            component.maintenance.extend(maintenance)

    def get_maintenance(
        self,
        age: Union[int, float],
        cycle_length: Union[int, float],
        horizon: Horizon,
        repeat: bool = True,
        include_history: bool = True,
        integers: bool = False,
    ) -> list[Maintenance]:
        """Get maintenance list for this strategy.

        Arguments:
            age: Starting age of a virtual component.
            cycle_length: Cycle length.
            horizon: Planning horizon to consider.
            repeat: Whether the cycle should be repeated until the end of the horizon.
            include_history: Whether to include historical maintenance entries for
                components that have a pre-defined age.
            overwrite: Whether to fully overwrite a component's maintenance planning
                with this new one or extend it.
            integers: Whether to force all times to be integers.

        Returns:
            Maintenance list.
        """
        assert len(self.tasks) == len(
            self.percentages
        ), "The number of tasks and percentages should be equal."

        start: Union[int, float] = -age
        end = horizon.end - horizon.start if horizon.end else cycle_length

        offsets = [p * cycle_length for p in self.percentages]
        tasks = sorted(zip(offsets, self.tasks), key=lambda x: x[0])

        maintenance = []
        cycles_offset: Union[int, float] = 0
        while start + cycles_offset < end:
            for offset, task in tasks:
                time = offset + cycles_offset + start
                if integers:
                    time = round(time)

                # No further planning beyond this point.
                if time > end:
                    break

                if not include_history and time < 0:
                    continue

                maintenance.append(
                    Maintenance(
                        f"{task.name} {time}",
                        task=task,
                        time=time,
                    )
                )

            # No further planning beyond this point.
            if not repeat or time > end:
                break

            cycles_offset = cycles_offset + cycle_length

        return maintenance



"""RaPlan distributions module."""

import math
import random
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Iterable, Optional, Union

from serde import InternalTagging, field, serde

__all__ = [
    "compound_probability",
    "Exponential",
    "LogNormal",
    "Pareto",
    "Triangular",
    "Uniform",
    "Weibull",
]


def compound_probability(probabilities: Iterable[float]) -> float:
    """Compound CDF value of multiple contributions."""
    return max(0.0, 1.0 - math.prod(1 - p for p in probabilities))


class Distribution(ABC):
    """Abstract base class for distributions."""

    def cdf(self, x: Union[int, float] = 1.0) -> float:
        """Cumulative probability density function of this distribution."""
        if x < 0:
            return 0.0
        return self._cdf(x)

    @abstractmethod
    def _cdf(self, x: Union[int, float] = 1.0) -> float:
        """Cumulative probability density function of this distribution."""

    @abstractmethod
    def sample(self) -> float:
        """Take a sample of this distribution."""


@serde(tagging=InternalTagging("type"))
@dataclass
class Exponential(Distribution):
    """Exponential distribution.

    Arguments:
        lambd: Occasion rate or 1 divided by the desired mean. In scheduling, it should
            be greater than zero.
    """

    lambd: Union[int, float] = 1.0

    def __post_init__(self):
        self.lambd = float(self.lambd)

    def _cdf(self, x: Union[int, float] = 1.0) -> float:
        return 1.0 - math.exp(-self.lambd * x)

    def sample(self) -> float:
        return random.expovariate(self.lambd)


@serde(tagging=InternalTagging("type"))
@dataclass
class LogNormal(Distribution):
    """Log-normal distribution.

    Arguments:
        mu: Mean of the underlying normal distribution. Any value.
        sigma: Standard deviation of the underlying normal distribution. Must be >0.
    """

    mu: Union[int, float] = 1.0
    sigma: Union[int, float] = 1.0

    def __post_init__(self):
        if self.sigma < 0.0:
            raise ValueError("sigma must be greater than 0.")

    def _cdf(self, x: Union[int, float] = 1.0) -> float:
        return 0.5 * (
            1.0 + math.erf(math.log(x - self.mu) / (self.sigma * math.sqrt(2)))
        )

    def sample(self) -> float:
        return random.lognormvariate(self.mu, self.sigma)


@serde(tagging=InternalTagging("type"))
@dataclass
class Pareto(Distribution):
    """Pareto distribution.

    Arguments:
        min_x: Minimum x value.
    """

    min_x: Union[int, float] = 1.0
    alpha: Union[int, float] = 1.0

    def __post_init__(self):
        try:
            assert self.min_x > 0, "min_x should be > 0."
            assert self.alpha > 0, "alpha should be > 0."
        except AssertionError as e:
            raise ValueError(e)

    def _cdf(self, x: Union[int, float] = 1.0) -> float:
        if x < self.min_x:
            return 0.0
        else:
            return 1 - (self.min_x / x) ** (self.alpha)

    def sample(self) -> float:
        return self.min_x * random.paretovariate(self.alpha)


@serde(tagging=InternalTagging("type"))
@dataclass
class Triangular(Distribution):
    """Triangular distribution.

    Samples are in the range [low, high] and with the specified mode between those
    bounds. The mode defaults to the midpoint between the two bounds.

    Arguments:
        low: Lower bound of the sample range.
        high: Upper bound of the sample range.
        mode: The mode between those two bounds.
    """

    low: Union[int, float] = 0.0
    high: Union[int, float] = 1.0
    mode: Optional[Union[int, float]] = field(default=None, skip_if_default=True)

    def __post_init__(self):
        if self.mode is None:
            self.mode = 0.5 * (self.high + self.low)

    def _cdf(self, x: Union[int, float] = 1.0) -> float:
        assert self.mode is not None
        a, b, c = self.low, self.high, self.mode
        if x < a:
            return 0.0
        elif x <= c:
            return (x - a) * (x - a) / ((b - a) * (c - a))
        elif x <= b:
            return 1.0 - (b - x) * (b - x) / ((b - a) * (b - c))
        else:
            return 1.0

    def sample(self) -> float:
        return random.triangular(self.low, self.high, self.mode)


@serde(tagging=InternalTagging("type"))
@dataclass
class Uniform(Distribution):
    """A uniform distribution."""

    a: Union[int, float] = 0.0
    b: Union[int, float] = 1.0

    def _cdf(self, x: Union[int, float] = 1.0) -> float:
        a, b = self.a, self.b
        if x <= a:
            return 0.0
        elif x < b:
            return (x - a) / (b - a)
        else:
            return 1.0

    def sample(self):
        return random.uniform(self.a, self.b)


@serde(tagging=InternalTagging("type"))
@dataclass
class Weibull(Distribution):
    """Weibull distribution (2-parameter).

    Arguments:
        alpha: Shape parameter.
        mtbf: Mean time between failure.
    """

    alpha: Union[int, float] = 2.0
    mtbf: Union[int, float] = 10.0

    @property
    def beta(self) -> float:
        """Weibull scale parameter."""
        return self.mtbf / math.gamma(1.0 + 1.0 / self.alpha)

    def _cdf(self, x: Union[int, float] = 1.0) -> float:
        return 1.0 - math.exp(-((x / self.beta) ** self.alpha))

    def sample(self) -> float:
        return random.weibullvariate(self.alpha, self.beta)


Distributions = Union[Exponential, LogNormal, Pareto, Triangular, Uniform, Weibull]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['raplan']

package_data = \
{'': ['*']}

install_requires = \
['pyserde[yaml]>=0.9.2,<0.10.0']

extras_require = \
{'all': ['plotly>=5.7.0,<6.0.0',
         'kaleido==0.2.1',
         'ragraph>=1.16.0,<2.0.0',
         'openpyxl>=3.0.10,<4.0.0'],
 'analysis': ['ragraph>=1.16.0,<2.0.0'],
 'excel': ['openpyxl>=3.0.10,<4.0.0'],
 'plot': ['plotly>=5.7.0,<6.0.0', 'kaleido==0.2.1']}

setup_kwargs = {
    'name': 'raplan',
    'version': '0.8.3',
    'description': 'Ratio planning and scheduling in Python.',
    'long_description': 'None',
    'author': 'Ratio Innovations B.V.',
    'author_email': 'info@ratio-case.nl',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://gitlab.com/ratio-case/python/raplan',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



"""RaPlan plotting module."""

from dataclasses import dataclass
from typing import Optional, Sequence, Union

from raplan.classes import Component, Horizon, Maintenance, Procedure, Project, System
from raplan.distributions import compound_probability

try:
    import plotly.graph_objects as go
    from plotly.colors import qualitative
    from plotly.subplots import make_subplots
except ImportError:  # pragma: nocover

    raise ImportError(
        "Please install plotting prerequisites by installing the 'plot' extra."
    )

_BAR_STYLE = dict(
    bargap=0.0,
    bargroupgap=0.1,
    barmode="relative",
)


@dataclass
class Compound:
    """A fictive compound of CFP carrying items."""

    subjects: Sequence[Union[Component, System, Project, "Compound"]]
    name: str = "Compound"

    def cfp(self, x: Union[int, float]) -> float:
        return compound_probability(c.cfp(x) for c in self.subjects)

    def get_ordered_maintenance(self) -> list[Maintenance]:
        return sorted(
            [m for s in self.subjects for m in s.get_ordered_maintenance()],
            key=lambda m: m.time,
        )


def get_overview_figure(
    subject: Union[
        Union[Component, System, Project, Compound],
        Sequence[Union[Component, System, Project, Compound]],
    ],
    xs: list[Union[int, float]],
    horizon: Optional[Horizon] = None,
    compound: Optional[str] = None,
    bar_width: str = "unit",
) -> go.Figure:
    """Get an overview figure consisting of a CFP plot, as well as cost and duration
    bar charts.

    Arguments:
        subject: Planning subject(s) with CFP and maintenance tasks.
        xs: Values to calculate the CFP at.
        horizon: Optional horizon to apply. It's start is used as the X-axis offset to
            apply and the end is used to then set the upper bound of the X-axis.
        compound: If not `None`, the title for a compound CFP line.
        bar_width: How to determine widths of bars shown. 'auto' for Plotly defaults or
            'unit' for widths of 1.

    Returns:
        Subplots overview figure.
    """
    fig = make_subplots(3, 1, shared_xaxes=True, x_title="Time")
    subjects = subject if isinstance(subject, list) else [subject]
    x_offset = 0.0 if horizon is None else horizon.start
    for s in subjects:
        fig.add_trace(_get_cfp_trace(s, xs, x_offset=x_offset), row=1, col=1)
    if compound and len(subjects) > 1:
        fig.add_trace(
            _get_cfp_trace(Compound(subjects, name=compound), xs, x_offset=x_offset)
        )
    for b in _get_barchart(
        Compound(subjects, "Cost"), prop="cost", x_offset=x_offset, bar_width=bar_width
    ):
        fig.add_trace(b, row=2, col=1)
    for b in _get_barchart(
        Compound(subjects, "Duration"),
        prop="duration",
        x_offset=x_offset,
        bar_width=bar_width,
    ):
        fig.add_trace(b, row=3, col=1)

    fig.layout.update(
        yaxis1=dict(title="CFP"),
        yaxis2=dict(title="Cost"),
        yaxis3=dict(title="Duration"),
        **_BAR_STYLE,
    )
    if horizon:
        _update_xaxis_horizon(fig, horizon)
    return fig


def get_cfp_figure(
    subject: Union[
        Union[Component, System, Project, Compound],
        Sequence[Union[Component, System, Project, Compound]],
    ],
    xs: list[Union[int, float]],
    horizon: Optional[Horizon] = None,
    compound: Optional[str] = None,
    thresholds: dict[str, float] = {"5%": 0.05},
) -> go.Figure:
    """Get a figure displaying the CFP or CFPs of the subject(s).

    Arguments:
        subject: Planning object(s) with a CFP method.
        xs: Time values to calculate CFP values at.
        horizon: Optional horizon to apply. It's start is used as the X-axis offset to
            apply and the end is used to then set the upper bound of the X-axis.
        compound: Whether to show a compound CFP line for the subjects given.
        thresholds: Threshold lines to show.

    Returns:
        Plotly figure.
    """
    subjects = subject if isinstance(subject, list) else [subject]
    x_offset = 0 if horizon is None else horizon.start
    cfps = [_get_cfp_trace(s, xs, x_offset=x_offset) for s in subjects]
    if compound and len(subjects) > 1:
        cfps.append(
            _get_cfp_trace(Compound(subjects, name=compound), xs, x_offset=x_offset)
        )
    threshold_lines = [
        go.Scatter(
            x=[xs[0] + x_offset, xs[-1] + x_offset],
            y=[v, v],
            name=k,
            line=dict(dash="dash", color="crimson"),
            mode="lines",
        )
        for k, v in thresholds.items()
    ]
    fig = go.Figure(
        cfps + threshold_lines,
        layout=dict(xaxis=dict(title="Time"), yaxis=dict(title="CFP")),
    )
    if horizon:
        _update_xaxis_horizon(fig, horizon)
    return fig


def get_cost_figure(
    subject: Union[
        Union[Component, System, Project, Compound],
        Sequence[Union[Component, System, Project, Compound]],
    ],
    horizon: Optional[Horizon],
    bar_width: str = "unit",
) -> go.Figure:
    """Get a cost bar chart figure.

    Arguments:
        subject: Planning object(s) with maintenance tasks with cost attached.
        horizon: Optional horizon to apply. It's start is used as the X-axis offset to
            apply and the end is used to then set the upper bound of the X-axis.
        bar_width: How to determine widths of bars shown. 'auto' for Plotly defaults or
            'unit' for widths of 1.

    Returns:
        Cost bar chart figure.
    """
    x_offset = 0 if horizon is None else horizon.start
    fig = go.Figure(
        _get_barchart(subject, x_offset=x_offset, prop="cost", bar_width=bar_width),
        layout=dict(xaxis=dict(title="Time"), yaxis=dict(title="Cost"), **_BAR_STYLE),
    )
    if horizon:
        _update_xaxis_horizon(fig, horizon)
    return fig


def get_duration_figure(
    subject: Union[
        Union[Component, System, Project, Compound],
        Sequence[Union[Component, System, Project, Compound]],
    ],
    horizon: Optional[Horizon] = None,
    bar_width: str = "unit",
) -> go.Figure:
    """Get a duration bar chart figure.

    Arguments:
        subject: Planning object(s) with maintenance tasks with duration attached.
        horizon: Optional horizon to apply. It's start is used as the X-axis offset to
            apply and the end is used to then set the upper bound of the X-axis.
        bar_width: How to determine widths of bars shown. 'auto' for Plotly defaults or
            'unit' for widths of 1.

    Returns:
        Duration bar chart figure.
    """
    x_offset = 0 if horizon is None else horizon.start
    fig = go.Figure(
        _get_barchart(subject, x_offset=x_offset, prop="duration", bar_width=bar_width),
        layout=dict(
            xaxis=dict(title="Time"), yaxis=dict(title="Duration"), **_BAR_STYLE
        ),
    )
    if horizon:
        _update_xaxis_horizon(fig, horizon)
    return fig


def _get_cfp_trace(
    subject: Union[Component, System, Project, Compound],
    xs: list[Union[int, float]],
    x_offset: Union[int, float] = 0.0,
) -> go.Scatter:
    """Get a line trace displaying the CFP of the subject.

    Arguments:
        subject: Planning object with a CFP method.
        xs: Time values to calculate CFP values at.
        x_offset: Value to offset x values with for display on axis.

    Returns:
        Plotly scatter trace.
    """
    ys = [subject.cfp(x) for x in xs]
    if x_offset != 0.0:
        # Redefine xs if offset is supplied.
        xs = [x + x_offset for x in xs]

    return go.Scatter(x=xs, y=ys, name=subject.name, mode="lines")


def _get_barchart(
    subject: Union[
        Union[Component, System, Project, Compound],
        Sequence[Union[Component, System, Project, Compound]],
    ],
    prop: str = "cost",
    x_offset: Union[int, float] = 0.0,
    bar_width: str = "unit",
) -> list[go.Bar]:
    subjects = subject if isinstance(subject, list) else [subject]
    bars = []
    for s in subjects:
        data: dict = dict(names=[], times=[], values=[])
        for m in s.get_ordered_maintenance():
            data["names"].append(m.name)
            data["times"].append(float(m.time + x_offset))
            data["values"].append(getattr(m.task, prop, 0.0))
        width: Optional[list[int]] = (
            len(data["times"]) * [1] if bar_width == "unit" else None
        )
        bars.append(
            go.Bar(
                x=data["times"],
                y=data["values"],
                hovertext=data["names"],
                name=s.name,
                width=width,
            ),
        )
    return bars


def _update_xaxis_horizon(fig: go.Figure, horizon: Horizon):
    lower = getattr(horizon, "start")
    upper = getattr(horizon, "end")
    xaxis = dict(range=(lower, upper))
    fig.update_layout(xaxis=xaxis)


def get_procedures_plot(procedures: list[Procedure]):
    """Get a procedures plot where systems are to be plotted versus standardized
    procedures (groups of tasks) and their duration and cost.

    Arguments:
        procedures: A list of procedures to be executed.
    """
    fig = make_subplots(
        rows=3,
        cols=1,
        shared_xaxes=True,
        vertical_spacing=0.1,
        row_heights=[0.6, 0.2, 0.2],
    )

    kinds = sorted(set(p.kind for p in procedures))
    kind_to_index = {value: i for i, value in enumerate(kinds)}
    _times = sorted(set(p.time for p in procedures))
    systems = sorted(set(p.system for p in procedures))
    system_to_index = {value: i for i, value in enumerate(systems)}

    procedures = sorted(
        procedures, key=lambda p: (p.kind, p.system, p.time, -p.cost, -p.duration)
    )

    colors = {
        kind: qualitative.Plotly[i % len(qualitative.Plotly)]
        for i, kind in enumerate(kinds)
    }
    hm_colors = [(float(i), colors[kind]) for i, kind in enumerate(kinds)]

    xs = list(range(int(_times[0]), int(_times[-1] + 1)))

    heatmap_z: list[list[Optional[int]]] = [[None for j in xs] for i in systems]
    heatmap_text: list[list[Optional[str]]] = [[None for j in xs] for i in systems]
    for p in procedures:
        row, col = system_to_index[p.system], int(p.time) - xs[0]
        heatmap_z[row][col] = kind_to_index[p.kind]
        heatmap_text[row][col] = p.kind

    for kind_index, kind in enumerate(kinds):
        times: list[Union[int, float]] = []
        costs: list[Union[int, float]] = []
        duras: list[Union[int, float]] = []
        objs: list[int] = []

        for p in procedures:
            if p.kind != kind:
                continue
            times.append(p.time)
            costs.append(p.cost)
            duras.append(p.duration)
            objs.append(system_to_index[p.system])

        # Heatmap, with adjusted colormaps to trick plotly's via legend toggles.
        hm_colors = [
            colors[kind] if kind == _kind else "rgba(0,0,0,0)" for _kind in kinds
        ]

        fig.add_trace(
            go.Heatmap(
                x=xs,
                y=systems,
                z=heatmap_z,
                legendgroup=kind,
                name="procedures",
                colorscale=hm_colors,
                showlegend=False,
                showscale=False,
                hoverongaps=False,
                xgap=2,
                ygap=2,
                hovertemplate="time: %{x}<br>system: %{y}<br>"
                + "procedure: %{customdata}",
                customdata=heatmap_text,
            )
        )

        # Cost
        fig.add_trace(
            go.Bar(
                x=times,
                y=costs,
                legendgroup=kind,
                name=kind,
                marker_color=colors[kind],
                hovertemplate="time: %{x}<br>cost: %{y}<br>object: %{customdata}",
                customdata=[systems[obj] for obj in objs],
            ),
            2,
            1,
        )

        # Duration
        fig.add_trace(
            go.Bar(
                x=times,
                y=duras,
                legendgroup=kind,
                name=kind,
                showlegend=False,
                marker_color=colors[kind],
            ),
            3,
            1,
        )

    fig.layout.update(
        xaxis=dict(showticklabels=True),
        xaxis2=dict(showticklabels=True),
        xaxis3=dict(title="Time"),
        yaxis1=dict(title="Subject"),
        yaxis2=dict(title="Cost"),
        yaxis3=dict(title="Duration"),
        barmode="stack",
        bargap=0.0,
        bargroupgap=0.1,
    )
    return fig
